/*
 *=BEGIN GPL
 *
 * This file is part of the SiteProbeQt program.
 *
 * Copyright(c) 2010 Darrell Duffy
 * http://www.windyweather.net/wp
 *
 * This file may be licensed under the terms of of the
 * GNU General Public License Version 2 (the ``GPL'').
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the GPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the GPL along with this
 * program. If not, go to http://www.gnu.org/licenses/gpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 *=END GPL
 */

#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "aboutdialog.h"
#include "viewlog.h"

#include <QDateTime>
#include <QDir>
#include <QTextStream>
#include <QtNetwork>
#include <QHostInfo>


MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
    ,m_bMonitoring(0)
    ,m_nTimer(0)
    ,m_pSettings(0)
    ,m_bScanActive(0)
{
    ui->setupUi(this);

    // Remove the Maximize button from the main window
    Qt::WindowFlags flags = 0;
    flags = windowFlags();
    flags = flags & (~Qt::WindowMaximizeButtonHint);
    setWindowFlags( flags );

    restoreSettings();

    // connect the tcp socket to our routines
    connect( &m_tcpSocket, SIGNAL(connected()), this, SLOT(siteConnected()));
    connect( &m_tcpSocket, SIGNAL(error(QAbstractSocket::SocketError)),
            this, SLOT(siteFailure(QAbstractSocket::SocketError)));

}


MainWindow::~MainWindow()
{
    delete ui;
}

//
// Settings files are .ini file format and are stored per user
// We store the site list, including the last state of the site and the last date/time checked.
//
#define SETTINGS_VERSION "101"
#define SETTINGS_VERSION_LAST "100"
void MainWindow::createSettings()
{
    // QSettings ( Format format, Scope scope, const QString & organization, const QString & application = QString(), QObject * parent = 0 )
    m_pSettings = new QSettings( QSettings::IniFormat, QSettings::UserScope,
                                  QString("WindyWeather"), QString("SiteProbeQtSiteList") );

    QString     sFile = m_pSettings->fileName();
    int         nidx = sFile.lastIndexOf( '/' ); //sFile.lastIndexOf( QDir::separator() );
    sFile = sFile.left(nidx) + QString("/SiteProbeQtLog.log");
    m_sLogPath = sFile;

}

//
// save Scan Interval in minutes and the site list with a count
//
void    MainWindow::saveSettings()
{
    createSettings();

    m_pSettings->beginGroup("WindowSettings");
    m_pSettings->setValue("size", this->size());
    m_pSettings->setValue("pos", this->pos());
    m_pSettings->endGroup();

    m_pSettings->beginGroup("ListColumns");
    m_pSettings->setValue("column0", ui->twSiteList->columnWidth( 0 ));
    m_pSettings->setValue("column1", ui->twSiteList->columnWidth( 1 ));
    m_pSettings->setValue("column2", ui->twSiteList->columnWidth( 2 ));
    m_pSettings->setValue("column3", ui->twSiteList->columnWidth( 3 ));
    m_pSettings->setValue("column4", ui->twSiteList->columnWidth( 4 ));
    m_pSettings->endGroup();

    m_pSettings->beginGroup("MonitoringParams");
    m_pSettings->setValue("version", SETTINGS_VERSION );
    m_pSettings->setValue("scan_minutes", ui->sbMinutes->value() );
    m_pSettings->endGroup();

    m_pSettings->beginGroup("SiteList");
    m_pSettings->setValue("version", SETTINGS_VERSION );

    int nItems = ui->twSiteList->topLevelItemCount();
    m_pSettings->setValue("site_count", nItems );

    for ( int i=0; i < nItems; i++ )
    {
        QTreeWidgetItem* pItem = ui->twSiteList->topLevelItem(i);

        QString key = QString("Site_%1").arg(i);
        m_pSettings->setValue( key, pItem->text(0) );
        key = QString("Status_%1").arg(i);
        m_pSettings->setValue( key, pItem->text(1) );
        key = QString("LastDateTime_%1").arg(i);
        m_pSettings->setValue( key, pItem->text(2) );

        // added for version 101
        key = QString("FirstFailure_%1").arg(i);
        m_pSettings->setValue( key, pItem->text(3) );
        key = QString("LastFailure_%1").arg(i);
        m_pSettings->setValue( key, pItem->text(4) );
    }

    m_pSettings->endGroup();
    delete m_pSettings;
    m_pSettings = NULL;

}

//
// restore everything we saved
//
void    MainWindow::restoreSettings()
{
    createSettings();

    // Get the size and position of the window as it is first created
    // as the default.
    QSize   defSize = this->size();
    QPoint  defPos = this->pos();

    m_pSettings->beginGroup("WindowSettings");
    this->resize(m_pSettings->value("size", defSize ).toSize());
    this->move(m_pSettings->value("pos", defPos ).toPoint());
    m_pSettings->endGroup();

    m_pSettings->beginGroup("ListColumns");
    ui->twSiteList->setColumnWidth( 0, m_pSettings->value( "column0", 100 ).toInt());
    ui->twSiteList->setColumnWidth( 1, m_pSettings->value( "column1", 50 ).toInt());
    ui->twSiteList->setColumnWidth( 2, m_pSettings->value( "column2", 60 ).toInt());
    ui->twSiteList->setColumnWidth( 3, m_pSettings->value( "column3", 60 ).toInt());
    ui->twSiteList->setColumnWidth( 4, m_pSettings->value( "column4", 60 ).toInt());
    m_pSettings->endGroup();

    qDebug("MainWindow::restoreSettings - columns restored");

    m_pSettings->beginGroup("MonitoringParams");
    QString versionParams = m_pSettings->value("version", QString(SETTINGS_VERSION) ).toString();

    ui->sbMinutes->setValue( m_pSettings->value("scan_minutes", 10 ).toInt() );
    m_pSettings->endGroup();

    m_pSettings->beginGroup("SiteList");

    QString versionSites = m_pSettings->value("version", QString(SETTINGS_VERSION) ).toString();

    ui->twSiteList->clear();

    int nItems = m_pSettings->value("site_count", 0 ).toInt();

    setStatus(QString("Restored %1 sites").arg(nItems));

    for ( int i=0; i < nItems; i++ )
    {
        QTreeWidgetItem* pItem = new QTreeWidgetItem();

        QString key = QString("Site_%1").arg(i);
        pItem->setText(0, m_pSettings->value( key, "" ).toString() );
        key = QString("Status_%1").arg(i);
        pItem->setText(1, m_pSettings->value( key, "" ).toString() );
        key = QString("LastDateTime_%1").arg(i);
        pItem->setText(2, m_pSettings->value( key, "" ).toString() );

        // we don't really need to do a version check on these since we just
        // default to blank.
        key = QString("FirstFailure_%1").arg(i);
        pItem->setText(3, m_pSettings->value( key, "" ).toString() );
        key = QString("LastFailure_%1").arg(i);
        pItem->setText(4, m_pSettings->value( key, "" ).toString() );

        ui->twSiteList->addTopLevelItem( pItem );
    }

    m_pSettings->endGroup();
    delete m_pSettings;
    m_pSettings = NULL;
    // Log when we startup
    logMessage(QString("Program start,,%1").arg(QDateTime::currentDateTime().toString("yyyy-MMM-dd hh:mm")));

}

// Show a status message at the bottom of the window
void MainWindow::setStatus( QString sts )
{
    ui->statusBar->showMessage( sts );
}

//
// Shutdown the program
//
// Log a message, close the log
// and save all the settings.
//
void MainWindow::on_actionQuit_triggered()
{
    // better way to close an application.
    close();
    //qApp->exit(); // just exit the application
}

void MainWindow::on_btnClose_clicked()
{
    on_actionQuit_triggered(); // We have things to do when we close
}

// Catch all ways to close and close the log, save settings on the way out.
void 	MainWindow::closeEvent ( QCloseEvent * event )
{
    qDebug("MainWindow::closeEvent");

    // Log the event, flush and Close the log file

    logMessage(QString("Program stop,,%1").arg(QDateTime::currentDateTime().toString("yyyy-MMM-dd hh:mm")));
    closeLog();
    saveSettings();

    QMainWindow::closeEvent( event );
}

//
// About Box is modal to keep it simple
//
void MainWindow::on_actionAbout_triggered()
{
    AboutDialog dialog(this);
    dialog.exec();              // show the About box as a modal dialog to keep it simple
}

// enable / disable GUI parts while we are monitoring
void MainWindow::enableGui( bool bEnable )
{
    ui->btnAddSite->setEnabled( bEnable );
    ui->btnRemoveSite->setEnabled( bEnable );
    ui->btnStartMonitoring->setEnabled( bEnable );
    ui->sbMinutes->setEnabled( bEnable );
    ui->leSiteName->setEnabled( bEnable );
}

//
// Add a site
//
void MainWindow::on_btnAddSite_clicked()
{

    if ( m_bMonitoring )
    {
        setStatus(tr("Stop Monitoring First"));
        return;
    }
    QString newSite;
    newSite = ui->leSiteName->text();

    if ( newSite.count() == 0 )
    {
        setStatus(tr("Enter a sitename first"));
        return;
    }

    // Don't insert duplicates
    QList<QTreeWidgetItem*> items = ui->twSiteList->findItems(newSite, Qt::MatchExactly, 0 );
    if ( items.count() )
    {
        setStatus(tr("Site is already in the list"));
        return;
    }

    // make sure we have a valid website name
    // This will not return failure if DNS is doing intercept pages.
    QHostInfo info = QHostInfo::fromName( newSite );
    QString serr = info.errorString();
    if ( info.error() )
    {
        setStatus(tr("Invalid site: %1 %2").arg(newSite).arg(serr) );
        return;
    }

    // Create the site entry
    QTreeWidgetItem* pitem = new QTreeWidgetItem();
    pitem->setText(0, newSite);
    pitem->setText(1, tr("Unknown"));
    pitem->setText(2, QDateTime::currentDateTime().toString("yyyy-MMM-dd hh:mm") );

    // insert as a toplevel item in the tree and set the status
    ui->twSiteList->insertTopLevelItem(0, pitem );
    setStatus(tr("Site %1 added").arg(newSite));
}

//
// Remove a site from the list
//
void MainWindow::on_btnRemoveSite_clicked()
{

    if ( m_bMonitoring )
    {
        setStatus(tr("Stop Monitoring First"));
        return;
    }
    QTreeWidgetItem *pitem = ui->twSiteList->currentItem();
    if ( pitem == NULL )
    {
        setStatus(tr("Select an item first"));
        return;
    }
    // Sites are simple enough to retype that we don't need to confirm
    // the removal.
    int idx = ui->twSiteList->indexOfTopLevelItem( pitem );
    if ( idx >= 0 )
    {
        pitem = ui->twSiteList->takeTopLevelItem(idx );
        delete pitem;
    }
    setStatus(tr("Site removed"));
}

//
// Start Monitoring
//
void MainWindow::on_btnStartMonitoring_clicked()
{
    if ( m_bMonitoring )
    {
        setStatus(tr("Monitoring running"));
        return;
    }

    if ( ui->twSiteList->topLevelItemCount() == 0 )
    {
        setStatus(tr("Add a site to the list first"));
        return;
    }

    // disable items during monitoring
    this->enableGui( false );

    setStatus(tr("Monitoring started"));
    logMessage(QString("Monitoring Started,,%1").arg(QDateTime::currentDateTime().toString("yyyy-MMM-dd hh:mm")));
    m_bMonitoring = true;
    m_bScanFirstTime = true;

    int minutes = ui->sbMinutes->value();
    if ( minutes <= 0 )
    {
        setStatus(tr("Monitoring at 1 minute intervals"));
        minutes = 1;
        ui->sbMinutes->setValue( 1 );
    }

#if 1 // normal case
    m_timer.start( minutes*60000, this );
#else
    // Test case - minutes equal seconds
    m_timer.start( minutes*1000, this );
#endif
    qDebug("MainWindow::btnStartMonitoring - scan started");
    // start the scan
    scanSites();
}

//
// Stop the scan
//
void MainWindow::on_btnStopMonitoring_clicked()
{

    if (! m_bMonitoring )
    {
        setStatus(tr("Monitoring not running"));
        return;
    }

    // Enable as we stop
    enableGui( true );

    // if the scan is active, then just hammer
    // it stopped and then clean up.
    if ( m_bScanActive )
    {
        stopScan();
    }
    logMessage(QString("Monitoring Stopped,,%1").arg(QDateTime::currentDateTime().toString("yyyy-MMM-dd hh:mm")));
    setStatus(tr("Monitoring stopped"));

    m_timer.stop();
    m_bMonitoring = false;
    qDebug("MainWindow::btnStopMonitoring - scan stopped");
}

//
// Timer Event
// Do a scan, or close the log
//
void MainWindow::timerEvent(QTimerEvent *event)
{
    // Monitoring timer
    if (event->timerId() == m_timer.timerId())
    {
        qDebug("MainWindow::timerEvent - start a scan");
        // scan the sites
        scanSites();
    }
    else if ( event->timerId() == m_logTimer.timerId() )
    {
        // Close the log if it has been open for a while.
        qDebug("MainWindow::timerEvent - close the log");
        // we need to close the log now
        closeLog();
    }
    else
    {
        // pass it on if its not for us
        QMainWindow::timerEvent(event);
    }
}

//
// View the log, but only if we are not monitoring
//
void MainWindow::on_actionOpen_Log_triggered()
{
    if ( m_bMonitoring )
    {
        setStatus(tr("Stop Monitoring First"));
        return;
    }

    setStatus( m_sLogPath );

    ViewLog dlg(this);

    dlg.setLogPath( m_sLogPath );
    dlg.exec();
}

//
// Log support
//
// Open the log in the settings area
//
void    MainWindow::openLog()
{
    // open the file if it's not already open
    if ( !m_fLogFile.isOpen() )
    {
        m_fLogFile.setFileName( this->m_sLogPath );
        if ( !m_fLogFile.open( QIODevice::Append | QIODevice::Text ) )
        {
            setStatus(tr("Difficulty opening log file"));
            return;
        }
    }
    // close the file after 2 minutes if we don't use it
    m_logTimer.stop();                      // stop any timer that's running
    m_logTimer.start( 2*60000, this );      // restart the timer each time
}

//
// Close the log file. Don't leave it open for long periods of time.
//
void    MainWindow::closeLog()
{
    if ( m_fLogFile.isOpen() )
    {
       m_fLogFile.close();
    }
    m_logTimer.stop();
}

//
// Log a string to the log
//
void    MainWindow::logMessage( QString str )
{
    // open the log, or just restart the timer each time
    openLog();

    QTextStream out(&m_fLogFile);
    out << str << "\n";
}


//
// A Test routine for the log file.
//
void MainWindow::on_actionTest_Log_File_triggered()
{
    openLog();
    QTextStream out(&m_fLogFile);

    int nItems = ui->twSiteList->topLevelItemCount();

    for ( int j=0; j < 10; j++ )
    {
        for ( int i=0; i < nItems; i++ )
        {
            QTreeWidgetItem* pItem = ui->twSiteList->topLevelItem(i);

            out << pItem->text(0) << ", ";
            out << pItem->text(1) << ", ";
            out << pItem->text(2);
            out << "\n";
        }
    }

    closeLog();
}

//
// scan the sites and update the data in the list
// and log any changes.
//
void    MainWindow::scanSites()
{
    qDebug("MainWindow::scanSites");
    if ( m_bScanActive )
    {
        qDebug("MainWindow::scanSites - scan still active");
        return; // scan already active, so just ignore this
    }

    qDebug("MainWindow::scanSites start scan");

    QString now = QDateTime::currentDateTime().toLocalTime().toString("yyyy-MMM-dd hh:mm:ss");
    setStatus(tr("Scan sites %1").arg(now));

    m_bStopScan = false;
    m_bScanActive = true;

    m_nSiteIndex = 0; // start with the first site

    nextSite();
}

//
// Stop the scan if it's active and we stop monitoring
// or we reach the last site in the list.
// The timer is still running.
//
void    MainWindow::stopScan()
{

    qDebug("MainWindow::stopScan");
    m_bStopScan = false;
    m_bScanActive = false;
    m_bScanFirstTime = false;
    // stop any ongoing activity
    m_tcpSocket.abort();
}

//
// Next site in the scan
//
void    MainWindow::nextSite()
{
    qDebug("MainWindow::nextSite");

    if ( m_bStopScan )
    {
        m_bScanActive = false;
        setStatus(tr("Scan stopped"));
        stopScan();
        return;
    }

    if ( m_nSiteIndex >= ui->twSiteList->topLevelItemCount() )
    {
        setStatus(tr("Scan complete"));
        stopScan();
        return;
    }

    // get the site name and start a scan.
    QTreeWidgetItem* pitem = ui->twSiteList->topLevelItem( m_nSiteIndex );
    if (!pitem)
    {
        setStatus(tr("Nasty - no item at %1").arg(m_nSiteIndex));
        stopScan();
        return;
    }
    QString site = pitem->text( 0 );
    QString now = QDateTime::currentDateTime().toString("yyyy-MMM-dd hh:mm");

    pitem->setText(2, now );
    setStatus(tr("Scanning %1 at %2").arg(site).arg(now));

    // do a DNS lookup on the site address
    QHostInfo   hostinfo;
    hostinfo = QHostInfo::fromName(site);

    // now set the host address
    QHostAddress host;
    QList<QHostAddress> addList;

    addList = hostinfo.addresses();
    qDebug("MainWindow::nextSite %d addresses", addList.count() );

    int adds = addList.count();

    // defend against no addresses found
    if ( adds )
    {
        host = addList[0];
    }

    if ( adds || host.isNull() )
    {
        //setStatus(tr("DNS lookup failiure for %1").arg(site));
        qDebug("MainWindow::nextSite - no address found");
    }
    setStatus(tr("site %1 : %2").arg(site).arg(host.toString()) );

    // stop any ongoing activity and try a new connection
    m_tcpSocket.abort();
    m_tcpSocket.connectToHost( host, 80 );
    // call backs will come to siteConnected or siteFailure after this
    // so we can just bugout.

    if ( !m_tcpSocket.isValid() )
    {
        qDebug("MainWindow::nextSite - socket not valid");
    }
    else
    {
        qDebug("MainWindow::nextSite - socket valid");
    }
}

//
// site connected, so just shut down the connection
// we know the site is there.
//
void MainWindow::siteConnected()
{
    qDebug("MainWindow::siteConnected");

    // get the site name and say we are up.
    QTreeWidgetItem* pitem = ui->twSiteList->topLevelItem( m_nSiteIndex );
    if (!pitem)
    {
        setStatus(tr("Nasty - no item at %1").arg(m_nSiteIndex));
        stopScan();
        return;
    }

    logSiteChange( pitem, tr("Up") );

    // stop any ongoing activity
    m_tcpSocket.abort();

    // next site in the list
    m_nSiteIndex++;
    nextSite();
}

//
// Site failed, so rack up the error and go to the next site.
//
void MainWindow::siteFailure(QAbstractSocket::SocketError socketError)
{
    qDebug("MainWindow::siteFailure");
    // get the site name and say we are up.
    QTreeWidgetItem* pitem = ui->twSiteList->topLevelItem( m_nSiteIndex );
    if (!pitem)
    {

        setStatus(tr("Nasty - no item at %1").arg(m_nSiteIndex));
        stopScan();
        return;
    }

    logSiteChange( pitem, tr("Down") );
    QString now = QDateTime::currentDateTime().toLocalTime().toString("yyyy-MMM-dd hh:mm");

    // If the FirstFailure column is null, then fill it
    // and always fill the LastFailure every time
    if ( pitem->text(3).count() == 0 )
    {
        pitem->setText(3, now);
    }
    pitem->setText(4, now );

    // stop any ongoing activity
    m_tcpSocket.abort();

    // Next site in the list
    m_nSiteIndex++;
    nextSite();
}

//
// Log a site if something has changed
//
void    MainWindow::logSiteChange( QTreeWidgetItem* pitem, QString sts )
{
    if ( m_bScanFirstTime || sts != pitem->text(1) )
    {
        //setStatus(tr("Site Change %1 %2").arg(pitem->text(0)).arg(sts));
        // change the status of the site
        pitem->setText(1, sts);
        logMessage(QString("%1,%2,%3").arg(pitem->text(0)).arg(pitem->text(1)).arg(pitem->text(2)) );
    }

}

void MainWindow::on_actionClear_All_Failures_triggered()
{
    if ( m_bMonitoring )
    {
        setStatus(tr("Stop Monitoring First"));
        return;
    }

    int nItems = ui->twSiteList->topLevelItemCount();

    for ( int i=0; i < nItems; i++ )
    {
        QTreeWidgetItem* pItem = ui->twSiteList->topLevelItem(i);
        pItem->setText(3, "" );
        pItem->setText(4, "" );
    }
    setStatus(tr("Failure date/times cleared"));
}

void MainWindow::on_actionClear_Selected_Failure_triggered()
{

    if ( m_bMonitoring )
    {
        setStatus(tr("Stop Monitoring First"));
        return;
    }
    QTreeWidgetItem *pitem = ui->twSiteList->currentItem();
    if ( pitem == NULL )
    {
        setStatus(tr("Select an item first"));
        return;
    }
    // Clear columns 3,4
    pitem->setText( 3, "" );
    pitem->setText( 4, "" );

    setStatus(tr("Failure date/time cleared"));
}

void MainWindow::on_actionTest_Failure_Columns_triggered()
{
    if ( m_bMonitoring )
    {
        setStatus(tr("Stop Monitoring First"));
        return;
    }

    int nItems = ui->twSiteList->topLevelItemCount();
    QString now = QDateTime::currentDateTime().toLocalTime().toString("yyyy-MMM-dd hh:mm");

    for ( int i=0; i < nItems; i++ )
    {
        QTreeWidgetItem* pItem = ui->twSiteList->topLevelItem(i);
        pItem->setText(3, now );
        pItem->setText(4, now );
    }
    setStatus(tr("Failure date/times set"));
}
